@@ -641,17 +641,28 @@ grid .help.listwin.sb -row 0 -column 1 -sticky ns -padx 1 -pady 1
 grid columnconfigure .help.listwin 0 -weight 1 -minsize 100
 grid rowconfigure .help.listwin 0 -weight 1 -minsize 100
 
-#--------------------------------------
-# Create a selection listbox
-#--------------------------------------
+#-----------------------------------------------------------------
+# Clear the selection listbox.  Create it if it does not exist.
+#-----------------------------------------------------------------
 
-toplevel .select -bg beige
-wm withdraw .select
+proc xcircuit::reset_select_listbox {} {
+   if {[catch {wm withdraw .select}]} {
+      toplevel .select -bg beige
+      wm withdraw .select
 
-listbox .select.sellist -bg white
-pack .select.sellist -expand true -fill both
-bind .select.sellist <ButtonRelease-1> {set idx [.select.sellist nearest %y]}
-bind .select.sellist <ButtonRelease-3> {set idx -1}
+      listbox .select.sellist -bg white
+      listbox .select.vallist -bg white
+      pack .select.sellist -side left -expand true -fill both
+      pack .select.vallist -side left -expand true -fill both
+      bind .select.sellist <ButtonRelease-1> {set p_key [.select.sellist get \
+		[.select.sellist nearest %y]]}
+      bind .select.sellist <ButtonRelease-3> {set p_key {}}
+      bind .select.vallist <ButtonRelease-3> {set p_key {}}
+   } else {
+      while {[.select.sellist size] > 0} {select.sellist delete 0}
+      while {[.select.vallist size] > 0} {select.vallist delete 0}
+   }
+}
 
 #--------------------------------------
 # Create the simple popup prompt window
@@ -724,27 +735,56 @@ proc xcircuit::printstring {stringlist} {
    return $p
 }
 
+proc xcircuit::promptmakeparam {} {
+   if {[xcircuit::select] > 0} {
+      .dialog.bbar.okay configure -command \
+	 {xcircuit::parameter make substring [.dialog.textent.txt get]; \
+	 wm withdraw .dialog}
+      .dialog.textent.title.field configure -text "Enter name for parameter"
+      .dialog.textent.txt delete 0 end
+      wm deiconify .dialog
+      focus .dialog.textent.txt
+   }
+}
+
 proc xcircuit::promptselectparam {} {
-   global idx
-   ::set dlist [xcircuit::parameter default substring]
-   ::set ilist {}
-   ::set idx -1
-   catch { ::set ilist [xcircuit::parameter get substring] }
+   global p_key
+   set dlist [xcircuit::parameter get substring]
+   set p_key {}
+
+   xcircuit::reset_select_listbox
    if {[llength $dlist] > 0} {
       if {[llength $dlist] == 1} {
-	 ::set idx 0
+	 set p_key [lindex [lindex $dlist 0] 0]
       } else {
          foreach i $dlist {
-            .select.sellist insert end [xcircuit::printstring $i]
+	    set p_name [lindex $i 0]
+	    set p_val [lindex $i 1]
+            .select.sellist insert end $p_name
+	    .select.vallist insert end [xcircuit::printstring $p_val]
          }
 	 wm deiconify .select
 	 raise .select
-         tkwait variable idx
+         tkwait variable p_key
 	 .select.sellist delete 0 end
+	 .select.vallist delete 0 end
 	 wm withdraw .select
       }
    }
-   return $idx
+   return $p_key
+}
+
+proc xcircuit::prompteditparams {} {
+   xcircuit::reset_select_listbox
+   set dlist [xcircuit::parameter get -forward]
+   if {[llength $dlist] > 0} {
+      foreach i $dlist {
+         .select.sellist insert end [lindex $i 0]
+	 .select.vallist insert end [lindex $i 1]
+      }
+      wm deiconify .select
+      raise .select
+   }
 }
 
 proc xcircuit::promptmakesymbol {} {
@@ -1297,7 +1337,7 @@ $m add cascade -label "Encoding" -menu $m.encodingmenu
 $m add cascade -label "Insert" -menu $m.insertmenu
 $m add cascade -label "Justification" -menu $m.justifymenu
 $m add command -label "Parameterize" \
-	-command {xcircuit::parameter make substring}
+	-command {xcircuit::promptmakeparam}
 $m add command -label "Unparameterize" \
 	-command {xcircuit::parameter forget substring}
 $m add separator
@@ -1346,6 +1386,7 @@ $m2 add command -label "1/2 space" -command "xcircuit::label insert halfspace"
 $m2 add command -label "1/4 space" -command "xcircuit::label insert quarterspace"
 $m2 add command -label "Kern" -command "xcircuit::promptkern"
 $m2 add command -label "Character" -command "xcircuit::label insert special"
+$m2 add command -label "Parameter" -command "xcircuit::label insert parameter"
 
 set m2 [menu $m.justifymenu -tearoff 0]
 $m2 add radio -label "Left Justified" -variable jhoriz -value left \
@@ -1607,10 +1648,6 @@ xcircuit::newpagebutton "Page 10" 10
 
 xcircuit::newlibrarybutton "User Library"
 
-# Automatic key focus on main drawing window
-
-bind .xcircuit.mainframe.mainarea.drawing <Enter> {focus %W}
-
 # New key bindings should pass through this function so that key
 # strokes are captured correctly for label editing.
 
@@ -1645,8 +1682,9 @@ bind .xcircuit.mainframe.mainarea.drawing <Key-Prior> {catch {page [expr {[page]
 
 # Wait for the drawing area to become visible, and set the focus on it.
 
+bind .xcircuit.mainframe.mainarea.drawing <Enter> {focus %W}
 tkwait visibility .xcircuit.mainframe.mainarea.drawing
-focus .xcircuit.mainframe.mainarea.drawing
+focus -force .xcircuit.mainframe.mainarea.drawing
 
 # This pops down the interpreter window, if the "console.tcl" script was run.
 
