@@ -8,11 +8,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
-#if defined(DARWIN)
-#include <sys/malloc.h>
-#else
-#include <malloc.h>
-#endif
 #include <sys/types.h>
 #include <errno.h>
 #include <limits.h>
@@ -2211,39 +2206,71 @@ void setfontmarks(short fvalue, short jvalue)
    }
 }
 
+/*----------------------------------------------*/
+/* GUI wrapper for startparam()			*/
+/*----------------------------------------------*/
+
+void promptparam(xcWidget w, caddr_t clientdata, caddr_t calldata)
+{
+   buttonsave *popdata = (buttonsave *)malloc(sizeof(buttonsave));
+
+   if (areastruct.selects == 0) return;  /* nothing was selected */
+
+   /* Get a name for the new object */
+
+   eventmode = NORMAL_MODE;
+   popdata->dataptr = NULL;
+   popdata->button = NULL; /* indicates that no button is assc'd w/ the popup */
+   popupprompt(w, "Enter name for new parameter:", "\0", stringparam, popdata, NULL);
+}
+
 #endif /* !TCL_WRAPPER */
 
 /*--------------------------------------------------------------*/
 /* Parameterize a label string (wrapper for parameterize()).	*/
+/* Assumes that the name has been generated by the popup prompt	*/
+/* and is therefore held in variable _STR2			*/
+/*--------------------------------------------------------------*/
+
+void stringparam(xcWidget w, caddr_t clientdata, caddr_t calldata)
+{
+   genericptr *settext;
+
+      if (eventmode == TEXT2_MODE || eventmode == TEXT3_MODE) {
+         settext = (genericptr *)EDITPART;
+         makeparam(TOLABEL(settext), _STR2);
+         objectdeselect();
+	 setparammarks(NULL);
+      }
+      else if (checkselect(LABEL)) parameterize(P_SUBSTRING, _STR2);
+      else
+         eventmode = LPARAM_MODE;
+}
+
+/*--------------------------------------------------------------*/
+/* Numerical parameterization (wrapper for parameterize()).	*/
 /*--------------------------------------------------------------*/
 
 void startparam(xcWidget w, pointertype value, caddr_t calldata)
 {
    genericptr *settext;
 
-   if (value == P_SUBSTRING) {
-      if (eventmode == TEXT2_MODE || eventmode == TEXT3_MODE) {
-         settext = (genericptr *)EDITPART;
-         makeparam(TOLABEL(settext));
-         objectdeselect();
-	 setparammarks(NULL);
-      }
-      else if (checkselect(LABEL)) parameterize((int)value);
-      else
-         eventmode = LPARAM_MODE;
+   if (value == (pointertype)P_SUBSTRING) {
+      strcpy(_STR2, (calldata != NULL) ? (char *)calldata : "substring");
+      stringparam(w, NULL, NULL);
    }
    else {
       if ((eventmode == NORMAL_MODE) && (areastruct.selects == 0)) { 
          eventmode = IPARAM_MODE;
          Wprintf("Click on object to parameterize.");
       }      
-      else parameterize((int)value);
+      else parameterize((int)value, NULL);
    }
 }
 
-/*--------------------------------------------------------------*/
-/* Unparameterize a label string (wrapper for parameterize()).	*/
-/*--------------------------------------------------------------*/
+/*---------------------------------------------------------------*/
+/* Unparameterize a label string (wrapper for unparameterize()). */
+/*---------------------------------------------------------------*/
 
 void startunparam(xcWidget w, pointertype value, caddr_t calldata)
 {
