@@ -11,11 +11,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#if defined(DARWIN)
-#include <sys/malloc.h>
-#else
-#include <malloc.h>
-#endif
 
 #include <X11/Intrinsic.h>
 #include <X11/StringDefs.h>
@@ -96,6 +91,142 @@ Widget *param_buttons[] = {
 };
 #endif
 
+/*----------------------------------------------------------------------*/
+/* Basic routines for matching parameters by key values.  Note that	*/
+/* this really, really ought to be replaced by a hash table search!	*/
+/*----------------------------------------------------------------------*/
+
+/*----------------------------------------------------------------------*/
+/* Check for the existance of a parameter with key "key" in object	*/
+/* "thisobj".   Return true if the parameter exists.			*/
+/*----------------------------------------------------------------------*/
+
+Boolean check_param(objectptr thisobj, char *key)
+{
+   oparamptr tops;
+
+   for (tops = thisobj->params; tops != NULL; tops = tops->next)
+      if (!strcmp(tops->key, key))
+	 return TRUE;
+
+   return FALSE;
+}
+
+/*----------------------------------------------------------------------*/
+/* Create a new parameter;  allocate memory for the parameter and the	*/
+/* key.									*/
+/*----------------------------------------------------------------------*/
+
+oparamptr make_new_parameter(char *key)
+{
+   oparamptr newops;
+
+   newops = (oparamptr)malloc(sizeof(oparam));
+   newops->next = NULL;
+   newops->key = (char *)malloc(1 + strlen(key));
+   strcpy(newops->key, key);
+   return newops;
+}
+
+/*----------------------------------------------------------------------*/
+/* Create a new element (numeric) parameter.  Fill in essential values	*/
+/*----------------------------------------------------------------------*/
+
+eparamptr make_new_eparam(char *key)
+{
+   eparamptr newepp;
+
+   newepp = (eparamptr)malloc(sizeof(eparam));
+   newepp->next = NULL;
+   newepp->key = (char *)malloc(1 + strlen(key));
+   strcpy(newepp->key, key);
+   newepp->pointno = 0;
+
+   return newepp;
+}
+
+/*----------------------------------------------------------------------*/
+/* Find the parameter in the indicated object by key			*/
+/*----------------------------------------------------------------------*/
+
+oparamptr match_param(objectptr thisobj, char *key)
+{
+   oparamptr fparam;
+
+   for (fparam = thisobj->params; fparam != NULL; fparam = fparam->next)
+      if (!strcmp(fparam->key, key))
+	 return fparam;
+
+   return NULL;		/* No parameter matched the key---error condition */
+}
+
+/*----------------------------------------------------------------------*/
+/* Find the parameter in the indicated instance by key.  If no such 	*/
+/* instance value exists, return NULL.					*/
+/*----------------------------------------------------------------------*/
+
+oparamptr match_instance_param(objinstptr thisinst, char *key)
+{
+   oparamptr fparam;
+
+   for (fparam = thisinst->params; fparam != NULL; fparam = fparam->next)
+      if (!strcmp(fparam->key, key))
+	 return fparam;
+
+   return NULL;		/* No parameter matched the key---error condition */
+}
+
+/*----------------------------------------------------------------------*/
+/* Find the parameter in the indicated instance by key.  If no such 	*/
+/* instance value exists, return the object (default) parameter.	*/
+/*----------------------------------------------------------------------*/
+
+oparamptr find_param(objinstptr thisinst, char *key)
+{
+   oparamptr fparam;
+   fparam = match_instance_param(thisinst, key);
+   if (fparam == NULL)
+      fparam = match_param(thisinst->thisobject, key);
+   return fparam;
+}
+
+/*----------------------------------------------------------------------*/
+/* Find the total number of parameters in an object			*/
+/*----------------------------------------------------------------------*/
+
+int get_num_params(objectptr thisobj)
+{
+   oparamptr fparam;
+   int nparam = 0;
+
+   for (fparam = thisobj->params; fparam != NULL; fparam = fparam->next)
+      nparam++;
+   return nparam;
+}
+
+/*----------------------------------------------------------------------*/
+/* Free an instance parameter.  Note that this routine does not free	*/
+/* any strings associated with string parameters!			*/
+/*----------------------------------------------------------------------*/
+
+void free_instance_param(objinstptr thisinst, oparamptr thisparam)
+{
+   oparamptr ops, lastops = NULL;
+
+   for (ops = thisinst->params; ops != NULL; ops = ops->next) {
+      if (ops == thisparam) {
+	 if (lastops != NULL)
+	    lastops->next = ops->next;
+	 else
+	    thisinst->params = ops->next;
+	 free(ops->key);
+	 free(ops);
+	 break;
+      }
+      lastops = ops; 
+   }
+}
+
 /*----------------------------------------------*/
 /* Draw a circle at all parameter positions	*/
 /*----------------------------------------------*/
@@ -104,12 +235,12 @@ void indicateparams(genericptr thiselem)
 {
    int i, j, k, which;
    oparamptr ops;
+   eparamptr epp;
 
    if (thiselem != NULL) {
-      for (i = 0; i < thiselem->num_params; i++) {
-         j = thiselem->passed[i].paramno;
-	 k = thiselem->passed[i].pointno;
-         ops = topobject->params[j];
+      for (epp = thiselem->passed; epp != NULL; epp = epp->next) { 
+	 ops = match_param(topobject, epp->key);
+	 k = epp->pointno;
 	 switch(ops->which) {
 	    case P_POSITION: case P_POSITION_X: case P_POSITION_Y:
 	       switch(thiselem->type) {
@@ -150,6 +281,7 @@ void setparammarks(genericptr thiselem)
 
    int i, j, paramno;
    oparamptr ops;
+   eparamptr epp;
 
    /* These match the order of parameter definitions in xcircuit.h */
    const char *paramvars[] = {"positionparam", "substringparam", "xparam",
@@ -163,9 +295,8 @@ void setparammarks(genericptr thiselem)
 
    /* For each parameter declared, set the corresponding Tcl variable */
    if (thiselem != NULL) {
-      for (i = 0; i < thiselem->num_params; i++) {
-         j = thiselem->passed[i].paramno;
-         ops = topobject->params[j];
+      for (epp = thiselem->passed; epp != NULL; epp = epp->next) {
+	 ops = match_param(topobject, epp->key);
          Tcl_SetVar(xcinterp, (char *)paramvars[ops->which], "true",
 		TCL_NAMESPACE_ONLY);
       }
@@ -180,6 +311,7 @@ void setparammarks(genericptr thiselem)
    const int rlength = sizeof(param_buttons) / sizeof(Widget *);
    int i, j, paramno;
    oparamptr ops;
+   eparamptr epp;
 
    /* Clear all checkmarks */
 
@@ -191,9 +323,8 @@ void setparammarks(genericptr thiselem)
    /* Check those properties which are parameterized in the element */
 
    if (thiselem != NULL) {
-      for (i = 0; i < thiselem->num_params; i++) {
-         j = thiselem->passed[i].paramno;
-         ops = topobject->params[j];
+      for (epp = thiselem->passed; epp != NULL; epp = epp->next) {
+         ops = match_param(topobject, epp->key);
          w = *param_buttons[ops->which];
          XtSetArg(wargs[0], XtNsetMark, True);
          XtSetValues(w, wargs, 1);
@@ -230,6 +361,7 @@ int opsubstitute(objectptr thisobj, objinstptr pinst)
    int i, j, k, ival;
    float fval;
    oparamptr ops, ips;
+   eparamptr epp;
    XPoint *setpt;
    Boolean needrecalc;	/* for arcs and splines */
    int retval = -1;
@@ -238,19 +370,19 @@ int opsubstitute(objectptr thisobj, objinstptr pinst)
 
       needrecalc = False;
       thiselem = *eptr;
-      if (thiselem->num_params == 0) continue;	/* Nothing to substitute */
+      if (thiselem->passed == NULL) continue;	/* Nothing to substitute */
 
-      for (i = 0; i < thiselem->num_params; i++) {
-         j = thiselem->passed[i].paramno;
-	 k = thiselem->passed[i].pointno;
-         if ((pinst != NULL) && (pinst->params != NULL))
-            ips = pinst->params[j];
-         else
-	    ips = NULL;
-         ops = thisobj->params[j];
+      for (epp = thiselem->passed; epp != NULL; epp = epp->next) {
+	 k = epp->pointno;
+	 ips = (pinst != NULL) ?  find_param(pinst, epp->key) : NULL;
+	 ops = match_param(thisobj, epp->key);
 
-	 ival = (ips != NULL) ? ips->parameter.ivalue : ops->parameter.ivalue;
-         fval = (ips != NULL) ? ips->parameter.fvalue : ops->parameter.fvalue;
+	 if (ips != NULL) {
+	    ival = ips->parameter.ivalue;
+	    fval = ips->parameter.fvalue;
+	 }
+	 else if (ops == NULL)
+	    continue;
 
          switch(ops->which) {
 	    case P_POSITION_X:
@@ -449,7 +581,7 @@ Boolean has_param(genericptr celem)
 	 if (cstr->type == PARAM_START)
 	    return TRUE;
    }
-   if (celem->num_params > 0) return TRUE;
+   if (celem->passed != NULL) return TRUE;
    return FALSE;
 }
 
@@ -482,19 +614,14 @@ void pwriteback(objinstptr thisinst)
    for (eptr = thisobj->plist; eptr < thisobj->plist + thisobj->parts; eptr++) {
 
       thiselem = *eptr;
-      if (thiselem->num_params == 0) continue;	/* Nothing to write back */
+      if (thiselem->passed == NULL) continue;	/* Nothing to write back */
 
-      for (i = 0; i < thiselem->num_params; i++) {
+      for (epp = thiselem->passed; epp != NULL; epp = epp->next) {
 
-         epp = thiselem->passed + i;
-	 j = epp->paramno;
 	 k = epp->pointno;
 
-         if ((pinst != NULL) && (pinst->params != NULL))
-            ips = pinst->params[j];
-         else
-	    ips = NULL;
-         ops = thisobj->params[j];
+         ips = (pinst != NULL) ? match_instance_param(pinst, epp->key) : NULL;
+         ops = match_param(thisobj, epp->key);
 
 	 ival = (ips != NULL) ? &ips->parameter.ivalue : &ops->parameter.ivalue;
 	 fval = (ips != NULL) ? &ips->parameter.fvalue : &ops->parameter.fvalue;
@@ -650,41 +777,26 @@ void replaceparams(objinstptr thisinst)
 
    thisobj = thisinst->thisobject;
 
-   for (i = 0; i < thisobj->num_params; i++) {
-      ops = thisobj->params[i];
-      ips = thisinst->params[i];
-      if (ips == NULL) {
-	 nullparms++;
-	 continue;  /* this parameter is already default */
-      }
+   for (ops = thisobj->params; ops != NULL; ops = ops->next) {
+      ips = match_instance_param(thisinst, ops->key);
+      if (ips == NULL) continue;  /* this parameter is already default */
 
       switch(ops->type) {
 	 case XC_STRING:
 	    if (stringcomp(ops->parameter.string, ips->parameter.string)) {
 	       freelabel(ops->parameter.string);
 	       ops->parameter.string = ips->parameter.string;
-	       free(ips);
-	       thisinst->params[i] = NULL;
-	       nullparms++;
+	       free_instance_param(thisinst, ips);
 	    }
 	    break;
 	 case XC_INT: case XC_FLOAT:
 	    if (ops->parameter.ivalue != ips->parameter.ivalue) {
 	       ops->parameter.ivalue = ips->parameter.ivalue;
-	       free(ips);
-	       thisinst->params[i] = NULL;
-	       nullparms++;
+	       free_instance_param(thisinst, ips);
 	    }
 	    break;
       }
    }
-
-   /* Object instance takes all default values---remove the params list */
-
-   if (nullparms == thisobj->num_params) {
-      free(thisinst->params);
-      thisinst->params = NULL;
-   }
 }
 
 /*------------------------------------------------------*/
@@ -699,7 +811,7 @@ void resolveparams(objinstptr thisinst)
    objectptr thisobj;
    liblistptr spec;
    oparamptr ops, ips;
-   int i, nullparms = 0;
+   int i;
 
    /* If the instance has no parameters itself, ignore it. */
    if (thisinst == NULL || thisinst->params == NULL) return;
@@ -724,40 +836,26 @@ void resolveparams(objinstptr thisinst)
 
    thisobj = thisinst->thisobject;
 
-   for (i = 0; i < thisobj->num_params; i++) {
-      ops = thisobj->params[i];
-      ips = thisinst->params[i];
-      if (ips == NULL) {
-	 nullparms++;
-	 continue;  /* this parameter is already default */
-      }
+   for (ops = thisobj->params; ops != NULL; ops = ops->next) {
+      ips = match_instance_param(thisinst, ops->key);
+      if (ips == NULL) continue;  /* this parameter is already default */
 
       switch(ops->type) {
 	 case XC_STRING:
 	    if (!stringcomp(ops->parameter.string, ips->parameter.string)) {
 	       freelabel(ips->parameter.string);
-	       free(ips);
-	       thisinst->params[i] = NULL;
-	       nullparms++;
+	       free_instance_param(thisinst, ips);
 	    }
 	    break;
 	 case XC_INT: case XC_FLOAT:
 	    if (ops->parameter.ivalue == ips->parameter.ivalue) {
-	       free(ips);
-	       thisinst->params[i] = NULL;
-	       nullparms++;
+	       free_instance_param(thisinst, ips);
 	    }
 	    break;
       }
    }
 
-   /* Object instance takes all default values---remove the params list */
-
-   if (nullparms == thisobj->num_params) {
-      free(thisinst->params);
-      thisinst->params = NULL;
-   }
-   else {
+   if (thisinst->params != NULL) {
       /* Object must recompute bounding box if any instance	*/
       /* uses a non-default parameter.				*/
 
@@ -768,51 +866,32 @@ void resolveparams(objinstptr thisinst)
 /*------------------------------------------------------*/
 /* Fill any NULL instance parameters with the values	*/
 /* from the calling instance, or from the instance	*/
-/* object's defaults if newinst = callinst.		*/ 
+/* object's defaults if destinst = sourceinst.		*/ 
 /*------------------------------------------------------*/
 
-void copyparams(objinstptr newinst, objinstptr callinst)
+void copyparams(objinstptr destinst, objinstptr sourceinst)
 {
-   objectptr callobj;
-   oparamptr *cparams;
-   int i;
+   oparamptr psource, cops, newops;
 
-   if (callinst == NULL) return;
+   if (sourceinst == NULL) return;
+   if (destinst == sourceinst)
+      psource = sourceinst->thisobject->params;
+   else
+      psource = sourceinst->params;
 
-   callobj = callinst->thisobject;
-   cparams = callinst->params;
-
-   if (callobj->num_params == 0) return;
-
-   if (newinst->params == NULL) {
-      newinst->params = (oparamptr *) malloc(callobj->num_params * sizeof(oparamptr));
-      for (i = 0; i < callobj->num_params; i++)
-	 newinst->params[i] = NULL;
-   }
-
-   for (i = 0; i < callobj->num_params; i++) {
-      if (newinst->params[i] == NULL) {
-         newinst->params[i] = (oparamptr) malloc(sizeof(oparam));
-         newinst->params[i]->type = callobj->params[i]->type;
-         newinst->params[i]->which = callobj->params[i]->which;
-         switch(callobj->params[i]->type) {
+   for (cops = psource; cops != NULL; cops = cops->next) {
+      if (match_instance_param(destinst, cops->key) == NULL) {
+	 newops = make_new_parameter(cops->key);
+	 newops->next = destinst->params;
+	 destinst->params = newops;
+	 newops->type = cops->type;
+	 newops->which = cops->which;
+         switch(cops->type) {
 	    case XC_STRING:
-	       if (newinst == callinst || callinst->params == NULL ||
-			callinst->params[i] == NULL)
-	          newinst->params[i]->parameter.string =
-	    	     	   stringcopy(callobj->params[i]->parameter.string);
-	       else
-	          newinst->params[i]->parameter.string =
-			   stringcopy(callinst->params[i]->parameter.string);
+	       newops->parameter.string = stringcopy(cops->parameter.string);
 	       break;
 	    case XC_INT: case XC_FLOAT:
-	       if (newinst == callinst || callinst->params == NULL ||
-			callinst->params[i] == NULL)
-	          newinst->params[i]->parameter.ivalue =
-			   callobj->params[i]->parameter.ivalue;
-	       else
-	          newinst->params[i]->parameter.ivalue =
-			   callinst->params[i]->parameter.ivalue;
+	       newops->parameter.ivalue = cops->parameter.ivalue;
 	       break;
 	    default:
 	       Fprintf(stderr, "Error:  bad parameter\n");
@@ -828,13 +907,27 @@ void copyparams(objinstptr newinst, objinstptr callinst)
 
 void makenumericalp(genericptr *gelem, u_int mode)
 {
-   oparamptr ops;
+   oparamptr ops, newops;
    eparamptr epp;
    XPoint *pptr;
+   char new_key[7];
+   int pidx;
+
+   /* Numerical parameters have designated keys to avoid the necessity	*/
+   /* of having to specify a key, to avoid conflicts with PostScript	*/
+   /* predefined keys, and other good reasons.				*/
+
+   /* The order of these type names must match the enumeration in xcircuit.h */
+
+   static char *param_keys[] = {
+	"p_gps", "p_str", "p_xps", "p_yps", "p_sty", "p_jst", "p_an1",
+	"p_an2", "p_rad", "p_axs", "p_rot", "p_scl", "p_wid", "p_col"
+   };
 
    /* Parameterized strings are handled by makeparam() */
+
    if ((*gelem)->type == LABEL && mode == P_SUBSTRING) {
-      makeparam(TOLABEL(gelem));
+      Fprintf(stderr, "Error: String parameter passed to makenumericalp()\n");
       return;
    }
 
@@ -846,117 +939,101 @@ void makenumericalp(genericptr *gelem, u_int mode)
 
    /* Make sure the parameter doesn't already exist.	   */
 
-   if ((*gelem)->num_params > 0) {
-      for (epp = (*gelem)->passed; epp < (*gelem)->passed
-		+ (*gelem)->num_params; epp++) {
-	 ops = *(topobject->params + epp->paramno);
-	 if (ops->which == (u_char)mode) {
-	    Fprintf(stderr, "Cannot duplicate a parameter!\n");
-	    return;
-	 }
+   for (epp = (*gelem)->passed; epp != NULL; epp = epp->next) {
+      ops = match_param(topobject, epp->key);
+      if (ops->which == (u_char)mode) {
+	 Fprintf(stderr, "Cannot duplicate a parameter!\n");
+	 return;
       }
    }
 
-   if (topobject->params == NULL) {
-      topobject->num_params = 1;
-      topobject->params = (oparamptr *)malloc(sizeof(oparamptr));
-   }
-   else {
-      topobject->num_params++;
-      topobject->params = (oparamptr *)realloc(topobject->params,
-		topobject->num_params * sizeof(oparamptr));
-   }
-   *(topobject->params + topobject->num_params - 1) = (oparamptr)
-	malloc(sizeof(oparam));
+   /* Ensure that no two parameters have the same name! */
 
-   initallinstances(topobject, topobject->num_params - 1);
+   strcpy(new_key, param_keys[mode]);
+   pidx = 0;
+   while (check_param(topobject, new_key)) {
+      pidx++;
+      sprintf(new_key, "%s%d", param_keys[mode], pidx);
+   }
 
-   ops = *(topobject->params + topobject->num_params - 1);
-   ops->type = XC_INT;		/* most commonly used value */
-   ops->which = (u_char)mode;	/* what kind of parameter */
+   newops = make_new_parameter(new_key);
+   newops->next = topobject->params;
+   topobject->params = newops;
+   newops->type = XC_INT;		/* most commonly used value */
+   newops->which = (u_char)mode;	/* what kind of parameter */
 
    /* Add the parameter to the element's parameter list */
 
-   if ((*gelem)->num_params == 0) {
-      (*gelem)->num_params = 1;
-      (*gelem)->passed = (eparamptr)malloc((*gelem)->num_params *
-		sizeof(eparam));
-   }
-   else {
-      (*gelem)->num_params++;
-      (*gelem)->passed = (eparamptr)realloc((*gelem)->passed,
-		(*gelem)->num_params * sizeof(eparam));
-   }
-   epp = (*gelem)->passed + (*gelem)->num_params - 1;
-   epp->paramno = (u_char)(topobject->num_params - 1);
-   epp->pointno = 0;
+   epp = make_new_eparam(new_key);
+   epp->next = (*gelem)->passed;
+   (*gelem)->passed = epp;
 
    switch((*gelem)->type) {
       case LABEL:
 	 switch(mode) {
 	    case P_POSITION_X:
-	       ops->parameter.ivalue = (int)TOLABEL(gelem)->position.x;
+	       newops->parameter.ivalue = (int)TOLABEL(gelem)->position.x;
 	       break;
 	    case P_POSITION_Y:
-	       ops->parameter.ivalue = (int)TOLABEL(gelem)->position.y;
+	       newops->parameter.ivalue = (int)TOLABEL(gelem)->position.y;
 	       break;
 	    case P_JUSTIFY:
-	       ops->parameter.ivalue = (int)TOLABEL(gelem)->justify;
+	       newops->parameter.ivalue = (int)TOLABEL(gelem)->justify;
 	       break;
 	    case P_ROTATION:
-	       ops->parameter.ivalue = (int)TOLABEL(gelem)->rotation;
+	       newops->parameter.ivalue = (int)TOLABEL(gelem)->rotation;
 	       break;
 	    case P_SCALE:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOLABEL(gelem)->scale;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOLABEL(gelem)->scale;
 	       break;
 	 }
 	 break;
       case ARC:
 	 switch(mode) {
 	    case P_POSITION_X:
-	       ops->parameter.ivalue = (int)TOARC(gelem)->position.x;
+	       newops->parameter.ivalue = (int)TOARC(gelem)->position.x;
 	       break;
 	    case P_POSITION_Y:
-	       ops->parameter.ivalue = (int)TOARC(gelem)->position.y;
+	       newops->parameter.ivalue = (int)TOARC(gelem)->position.y;
 	       break;
 	    case P_ANGLE1:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOARC(gelem)->angle1;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOARC(gelem)->angle1;
 	       break;
 	    case P_ANGLE2:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOARC(gelem)->angle2;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOARC(gelem)->angle2;
 	       break;
 	    case P_RADIUS:
-	       ops->parameter.ivalue = (int)TOARC(gelem)->radius;
+	       newops->parameter.ivalue = (int)TOARC(gelem)->radius;
 	       break;
 	    case P_MINOR_AXIS:
-	       ops->parameter.ivalue = (int)TOARC(gelem)->yaxis;
+	       newops->parameter.ivalue = (int)TOARC(gelem)->yaxis;
 	       break;
 	    case P_STYLE:
-	       ops->parameter.ivalue = (int)TOARC(gelem)->style;
+	       newops->parameter.ivalue = (int)TOARC(gelem)->style;
 	       break;
 	    case P_LINEWIDTH:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOARC(gelem)->width;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOARC(gelem)->width;
 	       break;
 	 }
 	 break;
       case OBJECT:
 	 switch(mode) {
 	    case P_POSITION_X:
-	       ops->parameter.ivalue = (int)TOOBJINST(gelem)->position.x;
+	       newops->parameter.ivalue = (int)TOOBJINST(gelem)->position.x;
 	       break;
 	    case P_POSITION_Y:
-	       ops->parameter.ivalue = (int)TOOBJINST(gelem)->position.y;
+	       newops->parameter.ivalue = (int)TOOBJINST(gelem)->position.y;
 	       break;
 	    case P_ROTATION:
-	       ops->parameter.ivalue = (int)TOOBJINST(gelem)->rotation;
+	       newops->parameter.ivalue = (int)TOOBJINST(gelem)->rotation;
 	       break;
 	    case P_SCALE:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOOBJINST(gelem)->scale;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOOBJINST(gelem)->scale;
 	       break;
 	 }
 	 break;
@@ -964,20 +1041,20 @@ void makenumericalp(genericptr *gelem, u_int mode)
 	 switch(mode) {
 	    case P_POSITION_X:
 	       pptr = TOPOLY(gelem)->points + areastruct.editcycle; 
-	       ops->parameter.ivalue = (int)pptr->x;
+	       newops->parameter.ivalue = (int)pptr->x;
 	       epp->pointno = areastruct.editcycle;
 	       break;
 	    case P_POSITION_Y:
 	       pptr = TOPOLY(gelem)->points + areastruct.editcycle; 
-	       ops->parameter.ivalue = (int)pptr->y;
+	       newops->parameter.ivalue = (int)pptr->y;
 	       epp->pointno = areastruct.editcycle;
 	       break;
 	    case P_STYLE:
-	       ops->parameter.ivalue = (int)TOPOLY(gelem)->style;
+	       newops->parameter.ivalue = (int)TOPOLY(gelem)->style;
 	       break;
 	    case P_LINEWIDTH:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOPOLY(gelem)->width;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOPOLY(gelem)->width;
 	       break;
 	 }
 	 break;
@@ -985,36 +1062,36 @@ void makenumericalp(genericptr *gelem, u_int mode)
 	 switch(mode) {
 	    case P_POSITION_X:
 	       pptr = TOSPLINE(gelem)->ctrl + areastruct.editcycle; 
-	       ops->parameter.ivalue = (int)pptr->x;
+	       newops->parameter.ivalue = (int)pptr->x;
 	       epp->pointno = areastruct.editcycle;
 	       break;
 	    case P_POSITION_Y:
 	       pptr = TOSPLINE(gelem)->ctrl + areastruct.editcycle; 
-	       ops->parameter.ivalue = (int)pptr->y;
+	       newops->parameter.ivalue = (int)pptr->y;
 	       epp->pointno = areastruct.editcycle;
 	       break;
 	    case P_STYLE:
-	       ops->parameter.ivalue = (int)TOSPLINE(gelem)->style;
+	       newops->parameter.ivalue = (int)TOSPLINE(gelem)->style;
 	       break;
 	    case P_LINEWIDTH:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOSPLINE(gelem)->width;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOSPLINE(gelem)->width;
 	       break;
 	 }
 	 break;
       case PATH:
 	 switch(mode) {
 	    case P_STYLE:
-	       ops->parameter.ivalue = (int)TOPATH(gelem)->style;
+	       newops->parameter.ivalue = (int)TOPATH(gelem)->style;
 	       break;
 	    case P_LINEWIDTH:
-	       ops->type = XC_FLOAT;
-	       ops->parameter.fvalue = TOPATH(gelem)->width;
+	       newops->type = XC_FLOAT;
+	       newops->parameter.fvalue = TOPATH(gelem)->width;
 	       break;
 	 }
 	 break;
    }
-
+   incr_changes(topobject);
 }
 
 /*--------------------------------------------------------------*/
@@ -1023,14 +1100,28 @@ void makenumericalp(genericptr *gelem, u_int mode)
 
 void unmakenumericalp(genericptr *gelem, u_int mode)
 {
+   eparamptr epp;
+   
    /* Parameterized strings are handled by makeparam() */
    if (((*gelem)->type == LABEL) && (mode == P_SUBSTRING)) {
       unparameterize(mode);
       setparammarks(*gelem);
       return;
    }
+   else {
 
-   /* else. . . (to be done) */
+      /* (to be done) */
+
+/*
+      for (epp = gelem->passed; epp != NULL; epp = epp->next)
+	 if (epp->which == (u_char)mode)
+*/
+	    /* remove the eparam */
+
+      /* check for any other references to the parameter.  If there	*/
+      /* are none, remove all instance records of the eparam, then	*/
+      /* remove the parameter itself from the object.			*/
+   }
 }
 
 /*--------------------------------------------------------------*/
@@ -1041,8 +1132,8 @@ void insertparam()
 {
    labelptr tlab;
    oparamptr ops;
-   short i, nparms = 0;
-   int result, selparm = -1;
+   int result, nparms;
+   char *selparm;
 
    /* Don't allow nested parameters */
 
@@ -1056,30 +1147,24 @@ void insertparam()
    /* If only one, then automatically use it.  Otherwise, prompt	*/
    /* for which parameter to use.					*/
 
-   for (i = 0; i < topobject->num_params; i++) {
-      ops = *(topobject->params + i);
-      if (ops->type == XC_STRING) {
+   for (ops = topobject->params; ops != NULL; ops = ops->next)
+      if (ops->type == XC_STRING)
 	 nparms++;
-	 selparm = i;
-      }
-   }
+
    if (nparms > 1) {
       char *newstr, *sptr;
-      char *sstart = (char *)malloc(250);
-      u_char *parammap = (u_char *)malloc(nparms);
+      char *sstart = (char *)malloc(1024);
       nparms = 0;
       strcpy(sstart, "Choose: ");
       sptr = sstart + 8;
-      for (i = 0; i < topobject->num_params; i++) {
-	 ops = *(topobject->params + i);
+      for (ops = topobject->params; ops != NULL; ops = ops->next) {
 	 if (ops->type == XC_STRING) {
-	    parammap[nparms] = i;
 	    nparms++;
 	    if (nparms != 1) {
 	       strcat(sptr, ", ");
 	       sptr += 2;
 	    }
-	    sprintf(sptr, "%d = <", nparms);
+	    sprintf(sptr, "%s = <", ops->key);
 	    newstr = stringprint(ops->parameter.string, NULL);
 	    strcat(sptr, newstr);
 	    free(newstr);
@@ -1099,36 +1184,42 @@ void insertparam()
 	 Tcl_SetResult(xcinterp, "Error in executing promptselectparam", NULL);
          return;
       }
-      result = Tcl_GetIntFromObj(xcinterp, Tcl_GetObjResult(xcinterp), &selparm);
-      if (result != TCL_OK) {
-	 Tcl_SetResult(xcinterp, "Bad parameter number.", NULL);
-	 selparm = -1;
+      selparm = Tcl_GetString(Tcl_GetObjResult(xcinterp));
+
+      ops = match_param(topobject, selparm);
+      if (ops == NULL)
+      {
+	 Tcl_SetResult(xcinterp, "No such parameter.", NULL);
+	 selparm = NULL;
+      }
+      else if (ops->type != XC_STRING)
+      {
+	 Tcl_SetResult(xcinterp, "The parameter is not a substring.", NULL);
+	 selparm = NULL;
       }
       /* Relate index in list (string parameters only) to index	*/
       /* over all parameters.					*/
 
-      if (selparm >= 0) selparm = parammap[selparm];
 #else
-      /* The non-Tcl version is a bit quirky, but it works. */
+      /* Need a new non-Tcl version method . . . */
+      ops = NULL;	/* placeholder */
 
-      selparm = parammap[getkeynum()];
 #endif
-      free(parammap);
    }
 
-   if ((selparm >= 0) && (selparm < topobject->num_params))
-      labeltext(PARAM_START, (char *)&selparm);
+   if (ops != NULL)
+      labeltext(PARAM_START, selparm);
    else
-      Wprintf("Parameter number nonexistent.");
+      Wprintf("No such parameter.");
 }
 
 /*--------------------------------------------------------------*/
 /* Parameterize a label string.					*/
 /*--------------------------------------------------------------*/
 
-void makeparam(labelptr thislabel)
+void makeparam(labelptr thislabel, char *key)
 {
-   oparamptr ops;
+   oparamptr newops;
    stringpart *begpart, *endpart;
 
    /* cannot form a parameter on a top-level page */
@@ -1167,104 +1258,49 @@ void makeparam(labelptr thislabel)
       endpart = makesegment(&thislabel->string, NULL);
    }
    begpart->type = PARAM_START;
-   begpart->data.paramno = (topobject->params == NULL) ? 0 :
-		topobject->num_params;
+   begpart->data.string = (char *)malloc(1 + strlen(key));
+   strcpy(begpart->data.string, key);
    endpart->type = PARAM_END;
 
    /* Now move the sections of string to the object parameter */
 
-   if (topobject->params == NULL) {
-      topobject->num_params = 1;
-      topobject->params = (oparamptr *)malloc(sizeof(oparamptr));
-   }
-   else {
-      topobject->num_params++;
-      topobject->params = (oparamptr *)realloc(topobject->params,
-		topobject->num_params * sizeof(oparamptr));
-
-      /* We have added a parameter.  This requires that all instances	*/
-      /* of the object which do not have NULL params must initialize	*/
-      /* and NULL the pointer to the new parameter.			*/
-
-      initallinstances(topobject, topobject->num_params - 1);
-   }
-   *(topobject->params + topobject->num_params - 1) = (oparamptr)
-	malloc(sizeof(oparam));
-
-
-   ops = *(topobject->params + topobject->num_params - 1);
-   ops->type = XC_STRING;
-   ops->which = P_SUBSTRING;
-   ops->parameter.string = begpart->nextpart;
+   newops = make_new_parameter(key);
+   newops->next = topobject->params;
+   topobject->params = newops;
+   newops->type = XC_STRING;
+   newops->which = P_SUBSTRING;
+   newops->parameter.string = begpart->nextpart;
    begpart->nextpart = endpart->nextpart;
    endpart->nextpart = NULL;
 
    textend = 0;
-}
-
-/*------------------------------------------------------*/
-/* Search and initialize the selected parameter in all	*/
-/* instances of the specified object.			*/
-/*------------------------------------------------------*/
-
-void initinst(objectptr topobj, objectptr refobj, short pnum)
-{
-   objinstptr tinst;
-   genericptr *pgen;
-
-   if (topobj == NULL) return;
-   for (pgen = topobj->plist; pgen < topobj->plist + topobj->parts; pgen++) {
-      if ((*pgen)->type == OBJECT) {
-	 tinst = TOOBJINST(pgen);
-	 if (tinst->thisobject == refobj) {
-	    if (tinst->params != NULL) {
-	       /* Make sure we've allocated enough space for this parameter */
-	       tinst->params = (oparamptr *)realloc(tinst->params,
-			tinst->thisobject->num_params * sizeof(oparamptr));
-	       *(tinst->params + pnum) = NULL;
-	    }
-	 }
-      }
-   }
+   incr_changes(topobject);
 }
 
 /*--------------------------------------------------------------*/
 /* Destroy the selected parameter in the indicated instance 	*/
 /*--------------------------------------------------------------*/
 
-void destroyinst(objinstptr tinst, objectptr refobj, short pnum)
+void destroyinst(objinstptr tinst, objectptr refobj, char *key)
 {
-   oparamptr *ops;
+   oparamptr ops;
    short k;
 
    if (tinst->thisobject == refobj) {
-      if (tinst->thisobject->num_params > 0) {
-	 if (tinst->params != NULL) {
-	    ops = tinst->params + pnum;
-	    if (*ops != NULL)
-	       freelabel((*ops)->parameter.string);
-	    free(*ops);
-	    *ops = NULL;
-	    if (pnum == 0) {
-	       free(tinst->params);
-	       tinst->params = NULL;
-	    }
-	    else {
-	       /* Re-order the remaining parameters */
-	       for (k = pnum; k < tinst->thisobject->num_params - 1; k++)
-	          *(tinst->params + k) = *(tinst->params + k + 1);
-	    }
-	 }
+      ops = match_instance_param(tinst, key);
+      if (ops != NULL) {
+	 freelabel(ops->parameter.string);
+	 free_instance_param(tinst, ops);
       }
    }
 }
-
+	
 /*--------------------------------------------------------------*/
 /* Search and destroy the selected parameter in all instances 	*/
 /* of the specified object.					*/
 /*--------------------------------------------------------------*/
 
-void searchinst(objectptr topobj, objectptr refobj, short pnum)
+void searchinst(objectptr topobj, objectptr refobj, char *key)
 {
    objinstptr tinst;
    genericptr *pgen;
@@ -1274,35 +1310,11 @@ void searchinst(objectptr topobj, objectptr refobj, short pnum)
    for (pgen = topobj->plist; pgen < topobj->plist + topobj->parts; pgen++) {
       if ((*pgen)->type == OBJECT) {
 	 tinst = TOOBJINST(pgen);
-	 destroyinst(tinst, refobj, pnum);
+	 destroyinst(tinst, refobj, key);
       }
    }
 }
 
-/*--------------------------------------------------------------*/
-/* Add a (NULL) parameter pointer to every instance of the	*/
-/* indicated object						*/
-/*--------------------------------------------------------------*/
-
-void initallinstances(objectptr thisobj, int p)
-{
-   int j, k;
-
-   for (k = 0; k < xobjs.pages; k++) {
-      if (xobjs.pagelist[k]->pageinst != NULL)
-         initinst(xobjs.pagelist[k]->pageinst->thisobject, thisobj, p);
-   }
-   for (j = 0; j < xobjs.numlibs; j++) {
-      for (k = 0; k < xobjs.userlibs[j].number; k++) {
-         initinst(*(xobjs.userlibs[j].library + k), thisobj, p);
-      }
-      initinst(xobjs.libtop[LIBRARY + j]->thisobject, thisobj, p);
-   }
-   for (k = 0; k < xobjs.delbuffer.number; k++) {
-      initinst(*(xobjs.delbuffer.library + k), thisobj, p);
-   }
-}
-
 /*--------------------------------------------------------------*/
 /* Check if this string contains a parameter			*/
 /*--------------------------------------------------------------*/
@@ -1327,7 +1339,8 @@ void unmakeparam(labelptr thislabel, stringpart *thispart)
    oparamptr ops;
    liblistptr spec;
    stringpart *strptr, *lastpart, *nextpart, *endpart, *newstr, *subs;
-   int p, j, k, l = -1;
+   char *key;
+   int j, k, l = -1;
    Boolean is_last = True;
    
    /* make sure there is a parameter here */
@@ -1336,7 +1349,7 @@ void unmakeparam(labelptr thislabel, stringpart *thispart)
       Wprintf("There is no parameter here.");
       return;
    }
-   p = thispart->data.paramno;
+   key = thispart->data.string;
 
    /* Unparameterizing can cause a change in the string */
    undrawtext(thislabel);
@@ -1352,7 +1365,7 @@ void unmakeparam(labelptr thislabel, stringpart *thispart)
 	 plab = TOLABEL(pgen);
 	 for (strptr = plab->string; strptr != NULL; strptr = strptr->nextpart) {
 	    if (strptr->type == PARAM_START) {
-	       if ((strptr != thispart) && (strptr->data.paramno == p)) {
+	       if ((strptr != thispart) && (!strcmp(strptr->data.string, key))) {
 		  is_last = False;
 		  break;
 	       }
@@ -1362,7 +1375,7 @@ void unmakeparam(labelptr thislabel, stringpart *thispart)
       }
    }
 
-   ops = *(topobject->params + p);
+   ops = match_param(topobject, key);
    subs = ops->parameter.string;
 
    /* Copy the default parameter into the place we are unparameterizing */
@@ -1403,54 +1416,30 @@ void unmakeparam(labelptr thislabel, stringpart *thispart)
 
    for (k = 0; k < xobjs.pages; k++) {
       if (xobjs.pagelist[k]->pageinst != NULL)
-         searchinst(xobjs.pagelist[k]->pageinst->thisobject, topobject, p);
+         searchinst(xobjs.pagelist[k]->pageinst->thisobject, topobject, key);
    }
    for (j = 0; j < xobjs.numlibs; j++) {
       for (k = 0; k < xobjs.userlibs[j].number; k++) {
 	 if (*(xobjs.userlibs[j].library + k) == topobject)
 	    l = j;
 	 else
-            searchinst(*(xobjs.userlibs[j].library + k), topobject, p);
+            searchinst(*(xobjs.userlibs[j].library + k), topobject, key);
       }
    }
    for (k = 0; k < xobjs.delbuffer.number; k++) {
-      searchinst(*(xobjs.delbuffer.library + k), topobject, p);
+      searchinst(*(xobjs.delbuffer.library + k), topobject, key);
    }
 
    /* Also check through all instances on the library page */
    if (l >= 0)
       for (spec = xobjs.userlibs[l].instlist; spec != NULL; spec = spec->next) 
-         destroyinst(spec->thisinst, topobject, p);
+         destroyinst(spec->thisinst, topobject, key);
 
+   free(ops->key);
    free(ops);
    
-   /* Re-order the remaining parameters */
-
-   for (k = p; k < topobject->num_params - 1; k++)
-      *(topobject->params + k) = *(topobject->params + k + 1);
-
-   /* Decrement the number of parameters */
-
-   if (--topobject->num_params == 0) {
-      free(topobject->params);
-      topobject->params = NULL;
-   }
-
-   /* decrement the parameter ID number of all parameters in all strings */
-   /* of this object whose parameter IDs are greater.			 */
-
-   for (pgen = topobject->plist; pgen < topobject->plist + topobject->parts;
-		pgen++) {
-      if ((*pgen)->type == LABEL) {
-	 plab = TOLABEL(pgen);
-	 for (strptr = plab->string; strptr != NULL; strptr = strptr->nextpart) {
-	    if (strptr->type == PARAM_START)
-	       if (strptr->data.paramno > p)
-		  strptr->data.paramno--;
-	 }
-      }
-   }
    freelabel(subs);
+   incr_changes(topobject);
 }
 
 /*------------------------------------------------------*/
@@ -1519,7 +1508,7 @@ void unparameterize(int mode)
 /* Wrapper for makeparam()					*/
 /*--------------------------------------------------------------*/
 
-void parameterize(int mode)
+void parameterize(int mode, char *key)
 {
    short *fselect, ptype;
    labelptr settext;
@@ -1537,7 +1526,7 @@ void parameterize(int mode)
       if ((mode == P_SUBSTRING) && (areastruct.selects == 1) &&
 		(SELECTTYPE(fselect) == LABEL)) {
          settext = SELTOLABEL(fselect);
-         makeparam(settext);
+         makeparam(settext, key);
       }
       else if (mode == P_POSITION) {
 	 makenumericalp(topobject->plist + (*fselect), P_POSITION_X);
@@ -1586,7 +1575,7 @@ int checklibtop()
    int i;
    pushlistptr thispush;
 
-   for (thispush = xobjs.stack; thispush != NULL; thispush = thispush->next)
+   for (thispush = areastruct.stack; thispush != NULL; thispush = thispush->next)
       if ((i = is_library(thispush->thisinst->thisobject)) >= 0)
 	 return i;
 
@@ -1598,10 +1587,18 @@ int checklibtop()
 /* (Reverts all parameters to default value)				*/
 /*----------------------------------------------------------------------*/
 
-void removeinst(objinstptr thisinst)
+void removeinstparams(objinstptr thisinst)
 {
-   free(thisinst->params);
-   thisinst->params = NULL;
+   oparamptr ops;
+
+   while (thisinst->params != NULL) {
+      ops = thisinst->params;
+      thisinst->params = ops->next;
+      free(ops->key);
+      if (ops->type == XC_STRING)
+	 freelabel(ops->parameter.string);
+      free(ops);
+   }
 }
 
 /*----------------------------------------------------------------------*/
@@ -1610,18 +1607,16 @@ void removeinst(objinstptr thisinst)
 
 void removeparams(objectptr thisobj)
 {
-   int i;
-   oparamptr pptr;
+   oparamptr ops;
 
-   for (i = 0; i < thisobj->num_params; i++) {
-      pptr = thisobj->params[i];
-      if (pptr->type == XC_STRING)
-	 freelabel(pptr->parameter.string);
-      free(pptr);
+   while (thisobj->params != NULL) {
+      ops = thisobj->params;
+      thisobj->params = ops->next;
+      free(ops->key);
+      if (ops->type == XC_STRING)
+	 freelabel(ops->parameter.string);
+      free(ops);
    }
-   free(thisobj->params);
-   thisobj->params = NULL;
-   thisobj->num_params = 0;
 }
 
 /*-----------------------------------------------------------------------*/
