@@ -524,19 +524,26 @@ extern void clearpins(objectptr);
 
 /* from parameter.c: */ 
 
+extern oparamptr match_param(objectptr, char *);
+extern oparamptr match_instance_param(objinstptr, char *);
+extern oparamptr find_param(objinstptr, char *);
+extern int get_num_params(objectptr);
+extern void free_instance_param(objinstptr, oparamptr);
+
+extern oparamptr make_new_parameter(char *);
+extern eparamptr make_new_eparam(char *);
+
 extern void indicateparams(genericptr);
 extern void makenumericalp(genericptr *, u_int);
 extern void unmakenumericalp(genericptr *, u_int);
 extern void noparmstrcpy(u_char *, u_char *);
 extern void insertparam();
-extern void makeparam(labelptr);
-extern void initinst(objectptr, objectptr, short);
-extern void searchinst(objectptr, objectptr, short);
-extern void initallinstances(objectptr, int);
+extern void makeparam(labelptr, char *);
+extern void searchinst(objectptr, objectptr, char *);
 extern stringpart *searchparam(stringpart *);
 extern void unmakeparam(labelptr, stringpart *);
 extern void unparameterize(int);
-extern void parameterize(int);
+extern void parameterize(int, char *);
 extern genericptr findparam(objectptr, void *, u_char);
 extern Boolean paramcross(objectptr, labelptr);
 extern oparamptr parampos(objectptr, labelptr, char *, short *, short *);
