@@ -36,6 +36,11 @@ typedef unsigned long	u_long;
 #define sign(a)			((a) <= 0 ? -1 : 1)
 #define max(a,b)		((a) < (b) ? (b) : (a))
 #define min(a,b)		((a) < (b) ? (a) : (b))
+
+/*----------------------------------------------------------------------*/
+/* Deal with 32/64 bit processors based on autoconf results.		*/
+/*----------------------------------------------------------------------*/
+
 #if SIZEOF_VOID_P == SIZEOF_UNSIGNED_INT
 #define Number(a)		(void *)((u_int) a)
 typedef u_int pointertype;
@@ -437,7 +442,6 @@ typedef struct _stringpart {
      u_char	*string;
      int	color;
      int	font;
-     int	paramno;
      float	scale;
      short	kern[2];
    }		data;
@@ -454,7 +458,13 @@ typedef struct _stringpart {
 
 enum paramtypes {XC_INT = 0, XC_FLOAT, XC_STRING};
 
-typedef struct {
+/* Object parameters: general key:value parameter model */
+/* Note that this really should be a hash table, not a linked list. . . */
+
+typedef struct _oparam *oparamptr;
+
+typedef struct _oparam {
+   char *	key;		/* name of the parameter */
    u_char	type;		/* type is from paramtypes list above */
    u_char	which;		/* what the parameter represents (P_*) */
    union {
@@ -462,12 +472,21 @@ typedef struct {
       int	  ivalue;	/* also covers type short int by typecasting */
       float	  fvalue;
    } parameter;			/* default or substitution value */
-} oparam, *oparamptr;
+   oparamptr	next;		/* next parameter in linked list */
+} oparam;
 
-typedef struct {
-   u_char	paramno;	/* parameter number to get value from.	   */
+/* Element parameters: reference back to the object's parameters */
+/* These parameters are forward-substituted when descending into */
+/* an object instance.						 */
+/* Note that this really should be a hash table, not a linked list. . . */
+
+typedef struct _eparam *eparamptr;
+
+typedef struct _eparam {
+   char *	key;		/* name of the parameter */
    u_char	pointno;	/* point number in point array, for polygons */
-} eparam, *eparamptr;
+   eparamptr	next;		/* next parameter in linked list */
+} eparam;
 
 /*----------------------------------------------------------------------*/
 /* Generic element type	is a superset of all elements.			*/	
@@ -499,12 +518,11 @@ typedef struct {
    u_char	type;
    int		color;
    eparamptr	passed;		/* numerical parameters passed from above */
-   u_char	num_params;
    XPoint	position;
    short	rotation;
    float	scale;
    objectptr	thisobject;
-   oparamptr    *params;	/* parameter substitutions for this instance */
+   oparamptr	params;		/* parameter substitutions for this instance */
    BBox		bbox;		/* per-instance bounding box information */
 #ifdef SCHEMA
    BBox		*schembbox;	/* Extra bounding box for pin labels */
@@ -549,7 +567,6 @@ typedef struct {
    u_char	type;
    int		color;
    eparamptr	passed;		/* numerical parameters passed from above */
-   u_char	num_params;
    XPoint	position;
    short	rotation;
    float	scale;
@@ -568,7 +585,6 @@ typedef struct {
    u_char	type;
    int		color;
    eparamptr	passed;		/* numerical parameters passed from above */
-   u_char	num_params;
    u_short	style;
    float	width;
    short	number;
@@ -583,7 +599,6 @@ typedef struct {
    u_char	type;
    int		color;
    eparamptr	passed;		/* numerical parameters passed from above */
-   u_char	num_params;
    u_short	style;
    float	width;
    XPoint	ctrl[4];
@@ -599,7 +614,6 @@ typedef struct {
    u_char	type;
    int		color;
    eparamptr	passed;		/* numerical parameters passed from above */
-   u_char	num_params;
    u_short	style;
    float	width;
    short	radius; 	/* x-axis radius */
@@ -620,7 +634,6 @@ typedef struct {
    u_char	type;
    int		color;
    eparamptr	passed;		/* numerical parameters passed from above */
-   u_char	num_params;
    u_short	style;
    float	width;
    short	parts;
@@ -730,8 +743,7 @@ typedef struct _xcobject {
 				/* parameterized elements) */
    short	parts;
    genericptr	*plist;	 	/* to be retypecast to label, polygon, etc. */
-   u_char	num_params;	/* number of calling parameters */
-   oparamptr	*params;	/* list of parameter defaults */
+   oparamptr	params;		/* list of parameters, with default values */
    
 #ifdef SCHEMA
    Highlight	highlight;	/* net to be highlighted on redraw */
@@ -899,7 +911,12 @@ typedef struct {
    Pixel	auxpix2, barpix2, parampix2;
 
    int		width, height, timeout;
-   XFontStruct	*xcfont, *helpfont, *filefont, *textfont, *titlefont;
+   XFontStruct	*helpfont, *filefont;
+
+#ifndef TCL_WRAPPER
+   XFontStruct	*xcfont, *textfont, *titlefont;
+#endif
+
 } ApplicationData, *ApplicationDataPtr;
 
 /*----------------------------------------------------------------------*/
@@ -986,8 +1003,9 @@ enum {
    XCF_Text_Return, XCF_Text_Delete, XCF_Text_Right, XCF_Text_Left,
    XCF_Text_Up, XCF_Text_Down, XCF_Text_Split,
    XCF_Text_Home, XCF_Text_End, XCF_Return_Char, XCF_Parameter,
-   XCF_Edit_Param, XCF_Edit_Break, XCF_Edit_Delete, XCF_Edit_Insert,
-   XCF_Edit_Next, XCF_Attach, XCF_Next_Library, XCF_Library_Directory,
+   XCF_Edit_Param,XCF_Edit_Break, XCF_Edit_Delete, XCF_Edit_Insert,
+   XCF_Edit_Append, XCF_Edit_Next, XCF_Attach,
+   XCF_Next_Library, XCF_Library_Directory,
    XCF_Library_Move, XCF_Library_Copy, XCF_Library_Edit, XCF_Library_Delete,
    XCF_Library_Duplicate, XCF_Library_Hide, XCF_Library_Virtual,
    XCF_Page_Directory, XCF_Library_Pop, XCF_Virtual,
